@@ -17,7 +17,9 @@ import requests
 import io
 import networkx as nx
 from shapely.geometry import LineString, Point, Polygon
-
+from rasterstats import zonal_stats
+from pathlib import Path
+from rasterio.mask import mask
 
 import warnings
 
@@ -32,10 +34,9 @@ from mapc_suitability.config import (
     mapc_lpd_folder,
     mass_mainland_crs,
     boston_assessors_csv,
-    boston_geojson_url,
+    boston_parcels_url,
 )
 
-
 ##### UNDERLYING FUNCTIONS #######
 def normalize_field(df, col: str):
     """
@@ -221,6 +222,28 @@ def get_file(dir_name: str, muni: str = None, fileType: str = None):
 
     return file
 
+# function to download from a link
+def get_gdf_from_zipped_link(url, file_type:str):
+    """
+
+    input: a URL that downloads a zipped shapefile, with file type (geojson, tif, shp)
+    output: a gdf of the download, with the shapefile download deleted (no storage need)
+
+    """
+    
+    path = download_and_extract_zip_to_temp(url)
+
+    # pull out just the file type of choice
+    shapefile = get_file(dir_name=path, fileType= ("." + file_type))
+
+    # read gdf from shapefile
+    gdf = gpd.read_file(shapefile)
+    gdf = gdf.to_crs(mass_mainland_crs)
+
+    # delete file after extracted
+    shutil.rmtree(path)
+
+    return gdf
 
 def get_most_updated_state_assessors_data(muni, path=None, include_row=False):
     """
@@ -231,19 +254,16 @@ def get_most_updated_state_assessors_data(muni, path=None, include_row=False):
         path = tempfile.mkdtemp()
 
     if muni == "Boston":
-        print("reading Boston data from ", boston_geojson_url, ". Update if necessary.")
+        print("reading Boston data from ", boston_parcels_url, ". Update if necessary.")
         # get geojson from boston's open data portal and download
 
         # unzip to temp folder
-        temp_folder = download_and_extract_zip_to_temp(boston_geojson_url)
-        layer = get_file(temp_folder, fileType="geojson")
-        boston_parcels_gdf = gpd.read_file(layer)
-        shutil.rmtree(temp_folder)
+        boston_parcels_gdf = get_gdf_from_zipped_link(boston_parcels_url, file_type="shp")
 
         # download csv
         s = requests.get(boston_assessors_csv).content
-        boston_assessors = pd.read_csv(io.StringIO(s.decode("utf-8")))
-        boston_assessors["GIS_ID"] = boston_assessors["GIS_ID"].astype(str)
+        boston_assessors = pd.read_csv(io.StringIO(s.decode("utf-8")), dtype={'GIS_ID': str})
+        #boston_assessors["GIS_ID"] = boston_assessors["GIS_ID"].astype(str)
 
         # merge gdf to assessor's data
         parcel_layer = boston_parcels_gdf.merge(
@@ -287,7 +307,10 @@ def get_most_updated_state_assessors_data(muni, path=None, include_row=False):
         # delete temporary directory
         shutil.rmtree(path)
 
+        parcel_layer = parcel_layer.to_crs(mass_mainland_crs)
+
         return parcel_layer
+    
 
 
 def get_landuse_data(muni):
@@ -347,39 +370,7 @@ def buffer_gdf(gdf, buffer_size, point=False):
     return gdf_buffer
 
 
-# function to download from a link
-def get_gdf_from_zipped_link(url, topic: str):
-    """
 
-    input: a URL that downloads a zipped shapefile
-    output: a gdf of the download, with the shapefile download deleted (no storage need)
-
-    """
-
-    # path = os.path.join(sp_data_dir, topic)
-    path = os.path.join(
-        "S:\\Network Shares\\K Drive\\DataServices\\Projects\\Current_Projects\\LandUse_VMT_GHG\\Analysis\\Spatial",
-        topic,
-    )
-    # path = os.path.join('K:\DataServices\Projects\Current_Projects\PDAs_PPAs\I90_PPA_PDA\Data', topic)
-    os.makedirs(path, exist_ok=True)
-
-    # extract zipped folder to path
-    with urlopen(url) as zipresp:
-        with zipfile.ZipFile(BytesIO(zipresp.read())) as zfile:
-            zfile.extractall(path)
-
-    # pull out just the .shp file
-    shapefile = get_file(dir_name=path, fileType=".shp")
-
-    # read gdf from shapefile
-    gdf = gpd.read_file(shapefile)
-    gdf = gdf.to_crs(mass_mainland_crs)
-
-    # delete file after extracted
-    shutil.rmtree(path)
-
-    return gdf
 
 
 def create_grid(feature, shape, side_length):
@@ -468,103 +459,6 @@ def create_grid(feature, shape, side_length):
     return grid
 
 
-def get_heat_score(heat_index_fp, muni_gdf):
-    """
-    For each census block  in the municipality, determines the relative heat index score compared to all other
-    block groups. Parcels within those block groups can then be prioritized higher.
-
-    Inputs: Muni boundary (gdf), heat index raster (geotiff), comparitive geography
-
-    """
-    from rasterstats import zonal_stats
-    from src.data.make_dataset import census_bl
-
-    with rasterio.open(heat_index_fp) as raster:
-        transform = raster.transform
-        lst = raster.read(1).astype("float64")
-
-    # retain census blocks within MUNI
-    census_bl["og_area"] = census_bl[
-        "geometry"
-    ].area  # get original area to do sliver analysis
-
-    muni_blocks = census_bl.clip(muni_gdf)
-    muni_blocks["pct_bg"] = (
-        (muni_blocks["geometry"].area) / (muni_blocks["og_area"])
-    ) * 100
-    muni_blocks["pct_bg"] = (
-        (muni_blocks["geometry"].area) / (muni_blocks["og_area"])
-    ) * 100
-    muni_blocks = muni_blocks.loc[
-        muni_blocks["pct_bg"] > 5
-    ].reset_index()  # only keep block groups where 5% or more of the bg remains. Reset index for zonal stats
-
-    ## ZONAL STATS ##
-    # run zonal stats on heat index for mmc - what is the mean lst index score across census block?
-    lst_stats = pd.DataFrame(
-        zonal_stats(muni_blocks, lst, affine=transform, stats="mean")
-    )
-
-    # join back to blocks, rename field
-    muni_blocks_heat = muni_blocks[["geoid20", "geometry"]].join(lst_stats)
-    muni_blocks_heat = muni_blocks_heat.rename(columns={"mean": "lst_mean"})
-
-    return muni_blocks_heat
-
-
-def get_open_space_density(muni_gdf):
-    from src.data.make_dataset import open_space_protected
-    from src.features.indicator_functions import overlap_stats
-
-    """
-    get a weighted (by open space area) density of protected open space over a grid
-
-    steps:
-    1. buffer grid cells
-    2. get average area of open space within [grid cells + buffer]
-    3. then join back to grid cells
-
-    """
-
-    ## STEP 1: BUFFER GRID CELLS ##
-
-    # create grid
-    muni_grid = create_grid(feature=muni_gdf, shape="hexagon", side_length=200)
-
-    # buffer grid cells
-    muni_grid_w_buffer = buffer_gdf(gdf=muni_grid, buffer_size=300)
-
-    ## STEP 2: GET AVERAGE AREA OF OPEN SPACE THAT OVERLAP BUFFERED GRID CELLS ##
-
-    # clip open space to buffered muni boundary
-    open_space_protected_muni = open_space_protected.clip(muni_grid_w_buffer)
-
-    # create field for area of open space
-    open_space_protected_muni["area"] = open_space_protected_muni.geometry.area
-
-    # run overlap statistics to get a mean size of overlapping open space for each grid cell
-    grid_openspace_overlap_mean = overlap_stats(
-        layer_1=muni_grid_w_buffer,
-        layer_2=open_space_protected_muni.drop(columns="LOC_ID"),
-        stats="mean",
-        stats_field="area",
-        new_field_name="os",
-        normalize=True,
-    ).fillna(0)
-
-    ## STEP 3: JOIN BACK TO ORIGINAL CELLS ##
-    muni_grid_w_open_space = (
-        muni_grid.merge(
-            grid_openspace_overlap_mean[["LOC_ID", "os_mean_n"]],
-            on="LOC_ID",
-            how="left",
-        )
-        .fillna(0)
-        .drop_duplicates()
-    )
-
-    return muni_grid_w_open_space
-
 
 def get_isochrones_from_input_poly(
     layer_1, layer_1_id, travel_times: list, travel_type: str
@@ -653,86 +547,95 @@ def get_isochrones_from_input_poly(
 ##### INDICATOR FUNCTIONS #######
 
 
-def if_overlap(
-    layer_1, layer_2, new_field_name: str, inverse: bool = False, point: bool = False
-):
-    """
-    For a given base layer (layer_1), identifies whether there is overlap (intersection) with a layer of interest (layer_2). If the layer of interest is a point layer,
-    identifies whether the point is within the base layer feature.
 
-    INPUTS:
-    - layer_1: (GeoDataFrame) The base layer being enriched
-    - layer_2: (GeoDataFrame) The overlap layer of interest
+def if_overlap (layer_1, 
+                layer_2,
+                new_field_name:str,
+                inverse:bool=False,
+                point:bool=False,
+                fields:list=None
+                ):
+   
+    '''
+    For a given base layer (layer_1), identifies whether there is overlap (intersection) with a layer of interest (layer_2). If the layer of interest is a point layer, 
+    identifies whether the point is within the base layer feature. 
+
+    INPUTS: 
+    - layer_1: (GeoDataFrame) The base layer being enriched  
+    - layer_2: (GeoDataFrame) The overlap layer of interest 
     - new_field_name: (string)  Input a string to represent this layer in the output dataset
 
     OPTIONAL PARAMETER(S):
-    - inverse: (bool, default False) When true, reverses the 0 and 1. This should be applied for layers in which overlap is unfavorable.
-    - point: (bool, default False) Select 'True' when layer_2 is a point layer. In these cases, the function looks for whether the base layer contains a point layer.
+    - inverse: (bool, default False) When true, reverses the 0 and 1. This should be applied for layers in which overlap is unfavorable. 
+    - point: (bool, default False) Select 'True' when layer_2 is a point layer. In these cases, the function looks for whether the base layer contains a point layer. 
+    - fields: (list) List of fields to retain from layer_2 
 
-    OUTPUT:
-    '[new_field_name]_ovlp' field added to layer_1
-        - Value of 1 if:
-            - There is overlap, and overlap is favorable (Inverse = 'False')
-            - There is no overlap, and overlap is unfavorable (Inverse = 'True')
-        - Value of 0 if:
-            - There is no overlap and overlap is favorable (Inverse = 'False')
-            - There is overlap is and overlap is unfavorable (Inverse = 'True')
-    """
+    OUTPUT: 
+    '[new_field_name]_ovlp' field added to layer_1 
+        - Value of 1 if: 
+            - There is overlap, and overlap is favorable (Inverse = 'False') 
+            - There is no overlap, and overlap is unfavorable (Inverse = 'True') 
+        - Value of 0 if: 
+            - There is no overlap and overlap is favorable (Inverse = 'False') 
+            - There is overlap is and overlap is unfavorable (Inverse = 'True') 
+    '''
 
-    # reproject all to mass mainland
+    #reproject all to mass mainland
     layer_1 = layer_1.to_crs(mass_mainland_crs)
     layer_2 = layer_2.to_crs(mass_mainland_crs)
 
-    # determine predicate based on 'point' input
+    #determine predicate based on 'point' input
     if point:
-        predicate = "contains"
+        predicate = 'contains'
     else:
-        predicate = "intersects"
+        predicate = 'intersects'
 
-    overlap_only = gpd.sjoin(
-        layer_1, layer_2, predicate=predicate, how="inner"
-    )  # only keep geometry from layer_1
 
-    ## MERGE BACK TO FULL DATASET ##
-    new_field_name = new_field_name + "_ovlp"
+    overlap_only = gpd.sjoin(layer_1, 
+                            layer_2,
+                            predicate=predicate,
+                            how='inner') #only keep geometry from layer_1
 
-    # add a field for whether it overlaps or not. If inverse is not selected, a value of 1 indicates overlap. Reverse is true if inverse not selected.
+    ## MERGE BACK TO FULL DATASET ## 
+    new_field_name = new_field_name + '_ovlp'
+
+    fields_list = [new_field_name]
+
+    #then add any other fields listed in input
+    if fields:
+        for field in fields:
+            fields_list.append(field) 
+
+    #add a field for whether it overlaps or not. If inverse is not selected, a value of 1 indicates overlap. Reverse is true if inverse not selected.
     if inverse:
         overlap_only[new_field_name] = 0
 
-        # join back to base layer
-        layer_1_with_overlap = (
-            layer_1.merge(
-                overlap_only[new_field_name],
-                how="left",
-                left_index=True,
-                right_index=True,
-            )
-            .fillna(1)
-            .drop_duplicates()
-        )
-
+        #join back to base layer 
+        layer_1_with_overlap = layer_1.merge(overlap_only[fields_list], 
+                                            how='left', 
+                                            left_index=True, 
+                                            right_index=True).fillna(1)
+        
+        layer_1_with_overlap = layer_1_with_overlap.loc[layer_1_with_overlap.astype(str).drop_duplicates().index]
+    
     else:
         overlap_only[new_field_name] = 1
 
-        # join back to base layer
-        layer_1_with_overlap = (
-            layer_1.merge(
-                overlap_only[new_field_name],
-                how="left",
-                left_index=True,
-                right_index=True,
-            )
-            .fillna(0)
-            .drop_duplicates()
-        )
+        #join back to base layer 
+        layer_1_with_overlap = layer_1.merge(overlap_only[fields_list], 
+                                            how='left', 
+                                            left_index=True, 
+                                            right_index=True).fillna(0)
 
-    return layer_1_with_overlap
+        layer_1_with_overlap = layer_1_with_overlap.loc[layer_1_with_overlap.astype(str).drop_duplicates().index]
+        
+    return(layer_1_with_overlap)
 
 
 def calculate_overlap(
     layer_1,
     layer_2,
+    id_field: str,
     how: str,
     new_field_name: str,
     normalize: bool = True,
@@ -749,6 +652,7 @@ def calculate_overlap(
     INPUTS:
     - layer_1: (GeoDataFrame, polygon) The base layer being enriched
     - layer_2: (GeoDataFrame, polygon or line) The overlap layer of interest
+    - id_field: (string) Input the field name for the ID field for layer 1
     - how: (string, default 'area') the type of calculation:
         - 'area' = calculates the area of overlap in meters squared
         - 'percent' = calculate the percentage of the base layer
@@ -768,7 +672,6 @@ def calculate_overlap(
         - '[new_field_name]_sqm_n' or '[new_field_name]_pct_n' or '[new_field_name]_m_n'
 
     """
-    from src.data.make_dataset import id_field
 
     if not how:
         how = "area"
@@ -905,11 +808,12 @@ def overlap_stats(
     layer_1,
     layer_2,
     new_field_name: str,
+    id_field: str,
     stats: str = "maj",
     normalize: bool = True,
     inverse: bool = False,
     stats_field: str = None,
-    nan_value=None,
+    nan_value=None
 ):
     """
     Assigns the base layer with the (max, min, mean, median, maj (majority), count, sum) value from an overlapping geography that may have multiple
@@ -918,6 +822,7 @@ def overlap_stats(
     INPUT PARAMETERS:
     - layer_1: (GeoDataFrame) The base layer being enriched
     - layer_2: (GeoDataFrame) The overlap layer of interest
+    - id_field: (str) Input the ID field for layer_1
     - new_field_name: (string) Input a string to represent this layer in the output dataset
 
     OPTIONAL PARAMETERS:
@@ -932,7 +837,6 @@ def overlap_stats(
     - [new_field_name]_['stat']
     - If normalize: '[new_field_name']_['stat']_n
     """
-    from src.data.make_dataset import id_field
 
     valid = {"max", "min", "mean", "median", "maj", "count", "sum"}
     if stats not in valid:
@@ -1023,6 +927,7 @@ def proximity(
     layer_1,
     layer_2,
     new_field_name: str,
+    id_field:str,
     normalize: bool = True,
     inverse: bool = False,
     unit: str = "m",
@@ -1035,6 +940,7 @@ def proximity(
     INPUT PARAMETERS:
     - layer_1: (GeoDataFrame) The base layer being enriched
     - layer_2: (GeoDataFrame) The layer of interest for proximity
+    - id_field: (string) Input the field name for the ID field for layer 1
     - new_field_name:(string) Input a string to represent this layer in the output dataset
 
 
@@ -1051,8 +957,6 @@ def proximity(
     - If fields list provided, those fields will also
     """
 
-    from src.data.make_dataset import id_field
-
     # first make sure units are correct. default to meters.
     if not unit:
         unit = "m"
@@ -1225,7 +1129,7 @@ def points_within_isochrone(
     ##CREATE DICTIONARY
     dict = {}
     dict[id_layer_2] = list
-    dict[new_field_name + "_count"] = sum
+    dict[new_field_name + "_count"] = "sum"
 
     fields_to_enrich = [id_layer_2, (new_field_name + "_count")]
 
@@ -1234,9 +1138,16 @@ def points_within_isochrone(
             dict[field] = list
         fields_to_enrich = fields_to_enrich + fields
 
-    # group by lot ID but aggregate with lists of
+    #group by ID but aggregate with lists of values in additional fields
     isochrone_grouped = layers_sjoin.groupby(by=id_layer_1).agg(dict).reset_index()
 
+    #for list fields, remove quotes and brackets, only keep unique items
+    fields_to_clean = fields_to_enrich.remove((new_field_name + "_count"))
+
+    for field in fields_to_clean:
+        isochrone_grouped[field] = isochrone_grouped[field].apply(lambda x: pd.Series(x).dropna().unique().tolist()).astype(
+            str).str.replace('[', '').str.replace(']', '').str.replace("'", '').str.lstrip(', ').str.rstrip(',').str.rstrip(', ')
+
     # join back to layer_1
     layer_1_enriched = layer_1.merge(
         isochrone_grouped, on=id_layer_1, how="left"
@@ -1318,7 +1229,7 @@ def points_within_buffer(
     dict = {}
     dict[id_layer_2] = list
     dict["distance"] = list
-    dict[new_field_name + "_count"] = sum
+    dict[new_field_name + "_count"] = "sum"
 
     fields_to_enrich = [id_layer_2, "distance", (new_field_name + "_count")]
 
@@ -1330,6 +1241,13 @@ def points_within_buffer(
     # group by lot ID but aggregate with lists of
     layer_1_grouped = layers_sjoin.groupby(id_layer_1, as_index=False).agg(dict)
 
+    #for list fields, remove quotes and brackets, only keep unique items
+    fields_to_clean = fields_to_enrich.remove((new_field_name + "_count"))
+
+    for field in fields_to_clean:
+        layer_1_grouped[field] = layer_1_grouped[field].apply(lambda x: pd.Series(x).dropna().unique().tolist()).astype(
+            str).str.replace('[', '').str.replace(']', '').str.replace("'", '').str.lstrip(', ').str.rstrip(',').str.rstrip(', ')
+
     # join back to layer_1
     layer_1_enriched = layer_1.merge(layer_1_grouped, on=id_layer_1, how="left").fillna(
         0
@@ -1344,6 +1262,7 @@ def calculate_indicator_score(
     function: str,
     layer_1,
     layer_2=None,
+    id_field=None,
     new_field_name=None,
     normalize=None,  # set normalize to True automatically?
     how=None,
@@ -1394,6 +1313,7 @@ def calculate_indicator_score(
             new_field_name=new_field_name,
             inverse=inverse,
             point=point,
+            fields=fields
         )
 
     elif function == "calculate_overlap":
@@ -1403,6 +1323,7 @@ def calculate_indicator_score(
         return calculate_overlap(
             layer_1=layer_1,
             layer_2=layer_2,
+            id_field=id_field,
             how=how,
             new_field_name=new_field_name,
             normalize=normalize,
@@ -1423,6 +1344,7 @@ def calculate_indicator_score(
             layer_1=layer_1,
             layer_2=layer_2,
             stats_field=stats_field,
+            id_field=id_field,
             new_field_name=new_field_name,
             stats=stats,
             normalize=normalize,
@@ -1437,6 +1359,7 @@ def calculate_indicator_score(
         return proximity(
             layer_1=layer_1,
             layer_2=layer_2,
+            id_field=id_field,
             new_field_name=new_field_name,
             normalize=normalize,
             inverse=inverse,
@@ -1472,3 +1395,72 @@ def calculate_indicator_score(
             id_layer_1=id_layer_1,
             id_layer_2=id_layer_2,
         )
+
+
+def run_zonal_stats(raster_fp, zone, stat='median', nodata=None, field_name=str):
+    '''
+    description of function here
+    '''
+    with rasterio.open(raster_fp) as raster:
+    #raster = rasterio.open(raster_fp)
+
+        #read in shapes, reset index to ensure join works at end
+        zone = zone.to_crs(raster.crs).reset_index(drop=True)
+        zone = zone[~(zone['geometry'].is_empty | zone['geometry'].isna())]
+
+        # Read the raster values
+        array = raster.read(1)
+        array = np.asarray(array, dtype=int)
+        #array[array < 0] = 0
+
+        # Get the affine
+        affine = raster.transform
+
+        #calculate zonal stats (mean only) and convert into a data frame
+        stats = zonal_stats(zone, array, affine=affine, stats=stat, nodata=nodata, all_touched=True)
+        stats_df = pd.DataFrame(stats)
+
+        #file name becomes column name
+        stats_df.rename(columns={stat: field_name}, inplace=True)
+
+        #add a field for percentile ranking? 
+        stats_df[('rnk_' + field_name)] = stats_df[field_name].rank(pct=True)
+        zonal_stats_df = zone.join(stats_df)
+
+    return(zonal_stats_df)
+
+
+#clip the raster to the muni
+
+def clip_raster(raster_fp, boundary):
+
+    '''
+    description of function here
+    '''
+
+    directory = os.path.dirname(raster_fp) #gets folder that raster is in
+
+    destination_fp = os.path.join(directory, 'clipped.tif') #sets a 
+
+    with rasterio.open(raster_fp) as src:
+        #update crs
+        
+        boundary = boundary.to_crs(src.crs)
+        src.meta.update({
+            'crs': src.crs
+            })
+        out_image, out_transform= mask(src, boundary.geometry,crop=True)
+        out_meta=src.meta.copy() # copy the metadata of the source DEM
+
+        
+    out_meta.update({
+        "driver":"Gtiff",
+        "height":out_image.shape[1], # height starts with shape[1]
+        "width":out_image.shape[2], # width starts with shape[2]
+        "transform":out_transform
+    })
+                
+    with rasterio.open(destination_fp,'w',**out_meta) as dst:
+        dst.write(out_image)
+
+    return destination_fp
\ No newline at end of file
